__author__ = "PyPSA meets Earth"
__copyright__ = "Copyright 2022, The PyPSA meets Earth Initiative"
__license__ = "MIT"

"""OSMPBF file reader.

This module provides a reader for OpenStreetMap PBF files.

Modified from esy-osm-pbf/file.py

"""

import struct
import zlib
from itertools import accumulate

from . import Node, Relation, Way, fileformat_pb2, osmformat_pb2


def decode_strmap(primitive_block):
    """
    Decode the string table from a primitive block into a list of strings.
    """

    return tuple(s.decode('utf8') for s in primitive_block.stringtable.s)


def iter_blocks(file):
    """
    Iterate over the blocks in a OpenStreetMap PBF file.
    """

    ofs = 0
    while True:
        file.seek(ofs)
        data = file.read(4)
        if len(data) < 4:
            return
        header_size, = struct.unpack('>I', data)
        header = fileformat_pb2.BlobHeader()
        header.ParseFromString(file.read(header_size))
        ofs += 4 + header_size
        yield ofs, header
        ofs += header.datasize


def read_blob(file, ofs, header):
    """
    Read a blob from a OpenStreetMap PBF file.
    """

    file.seek(ofs)
    blob = fileformat_pb2.Blob()
    blob.ParseFromString(file.read(header.datasize))
    if blob.raw:
        return blob.raw
    elif blob.zlib_data:
        return zlib.decompress(blob.zlib_data)
    else:
        raise ValueError('Unknown blob type')


def parse_tags(strmap, keys_vals):
    """
    Parse the tags from a OpenStreetMap PBF file.
    """

    tags = {}
    is_value = False
    for idx in keys_vals:
        if idx == 0:
            yield tags
            tags = {}
        elif is_value:
            tags[key] = strmap[idx]
            is_value = False
        else:
            key = strmap[idx]
            is_value = True


def iter_primitive_block(primitive_block):
    """
    Iterate over the elements in a primitive block.
    """

    strmap = decode_strmap(primitive_block)
    for group in primitive_block.primitivegroup:
        for id, tags, lonlat in iter_nodes(primitive_block, strmap, group):
            yield Node(id, tags, lonlat)

        for id, refs, tags in iter_ways(primitive_block, strmap, group):
            yield Way(id, tags, refs)

        for id, members, tags in iter_relations(primitive_block, strmap, group):
            yield Relation(id, tags, members)


def iter_nodes(block, strmap, group):
    dense = group.dense
    if not dense:
        raise ValueError('Only dense nodes are supported')
    granularity = block.granularity or 100
    lat_offset = block.lat_offset or 0
    lon_offset = block.lon_offset or 0
    coord_scale = 0.000000001
    id = lat = lon = tag_pos = 0
    for did, dlat, dlon, tags in zip(
            dense.id, dense.lat, dense.lon,
            parse_tags(strmap, dense.keys_vals)):
        id += did
        lat += coord_scale * (lat_offset + granularity * dlat)
        lon += coord_scale * (lon_offset + granularity * dlon)
        yield (id, tags, (lon, lat))


def iter_ways(block, strmap, group):
    for way in group.ways:
        tags = {strmap[k]: strmap[v] for k, v in zip(way.keys, way.vals)}
        refs = tuple(accumulate(way.refs))
        yield way.id, refs, tags


def iter_relations(block, strmap, group):
    namemap = {}
    for relation in group.relations:
        tags = {
            strmap[k]: strmap[v] for k, v in zip(relation.keys, relation.vals)
        }
        refs = tuple(accumulate(relation.memids))
        members = [
            (
                ref,
                namemap.setdefault(
                    rel_type, osmformat_pb2.Relation.MemberType.Name(rel_type)
                ),
                strmap[sid],
            )
            for ref, rel_type, sid in zip(
                refs, relation.types, relation.roles_sid
            )
        ]

        yield relation.id, members, tags



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: earth_osm/osmpbf/fileformat.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n!earth_osm/osmpbf/fileformat.proto\x12\x06OSMPBF\"\xa5\x01\n\x04\x42lob\x12\x10\n\x08raw_size\x18\x02 \x01(\x05\x12\r\n\x03raw\x18\x01 \x01(\x0cH\x00\x12\x13\n\tzlib_data\x18\x03 \x01(\x0cH\x00\x12\x13\n\tlzma_data\x18\x04 \x01(\x0cH\x00\x12!\n\x13OBSOLETE_bzip2_data\x18\x05 \x01(\x0c\x42\x02\x18\x01H\x00\x12\x12\n\x08lz4_data\x18\x06 \x01(\x0cH\x00\x12\x13\n\tzstd_data\x18\x07 \x01(\x0cH\x00\x42\x06\n\x04\x64\x61ta\"?\n\nBlobHeader\x12\x0c\n\x04type\x18\x01 \x02(\t\x12\x11\n\tindexdata\x18\x02 \x01(\x0c\x12\x10\n\x08\x64\x61tasize\x18\x03 \x02(\x05\x42\x0f\n\rcrosby.binary')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'earth_osm.osmpbf.fileformat_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\rcrosby.binary'
  _BLOB.fields_by_name['OBSOLETE_bzip2_data']._options = None
  _BLOB.fields_by_name['OBSOLETE_bzip2_data']._serialized_options = b'\030\001'
  _BLOB._serialized_start=46
  _BLOB._serialized_end=211
  _BLOBHEADER._serialized_start=213
  _BLOBHEADER._serialized_end=276
# @@protoc_insertion_point(module_scope)



__author__ = "PyPSA meets Earth"
__copyright__ = "Copyright 2022, The PyPSA meets Earth Initiative"
__license__ = "MIT"

"""Utilities functions for earth_osm

This module contains utilities functions for handling OSM data.

"""


import logging
import os
import pandas as pd
import geopandas as gpd
from shapely.geometry import LineString, Point, Polygon

from earth_osm.config import primary_feature_element

logger = logging.getLogger("osm_data_extractor")
logger.setLevel(logging.INFO)

# geo_crs: EPSG:4326  # general geographic projection, not used for metric measures. "EPSG:4326" is the standard used by OSM and google maps
# distance_crs: EPSG:3857  # projection for distance measurements only. Possible recommended values are "EPSG:3857" (used by OSM and Google Maps)
# area_crs: ESRI:54009  # projection for area measurements only. Possible recommended values are Global Mollweide "ESRI:54009"

def lonlat_lookup(df_way, primary_data):
    """
    Lookup refs and convert to list of longlats
    """
    if "refs" not in df_way.columns:
        logger.warning("refs column not found")

    def look(ref):
        lonlat_row = list(map(lambda r: tuple(primary_data["Node"][str(r)]["lonlat"]), ref))
        return lonlat_row

    lonlat_list = df_way["refs"].apply(look)

    return lonlat_list

def way_or_area(df_way):
    if "refs" not in df_way.columns:
        logger.warning("refs column not found")
    
    def check_closed(ref):
        if (ref[0] == ref[-1]) and (len(ref) >= 3):
            return "area"
        elif len(ref) >= 3:
            return "way"
        else:
            logger.debug(f"Length of ref {len(ref)}")
            return pd.nan


    type_list = df_way["refs"].apply(check_closed)

    return type_list


def convert_ways_points(df_way, primary_data):
    """
    Convert Ways to Point Coordinates
    """
    lonlat_list = lonlat_lookup(df_way, primary_data)
    way_polygon = list(
        map(
            lambda lonlat: Polygon(lonlat) if len(lonlat) >= 3 else Point(lonlat[0]),
            lonlat_list,
        )
    )
    area_column = list(
        map(
            int,
            round(
                gpd.GeoSeries(way_polygon)
                .set_crs("EPSG:4326")
                .to_crs("ESRI:54009")
                .area,
                -1,
            ),
        )
    )  # TODO: Rounding should be done in cleaning scripts

    def find_center_point(p):
        if p.geom_type == "Polygon":
            center_point = p.centroid
        else:
            center_point = p
        return list((center_point.x, center_point.y))

    lonlat_column = list(map(find_center_point, way_polygon))

    # df_way.drop("refs", axis=1, inplace=True, errors="ignore")
    df_way.insert(0, "Area", area_column)
    df_way.insert(0, "lonlat", lonlat_column)


def convert_ways_polygons(df_way, primary_data):
    """
    Convert Ways to Polygon and Point Coordinates
    """
    lonlat_list = lonlat_lookup(df_way, primary_data)
    way_polygon = list(
        map(
            lambda lonlat: Polygon(lonlat) if len(lonlat) >= 3 else Point(lonlat[0]),
            lonlat_list,
        )
    )


    df_way.insert(0, "Area", area_column)

def convert_ways_lines(df_way, primary_data):
    """
    Convert Ways to Line Coordinates

    Args:

    """
    lonlat_list = lonlat_lookup(df_way, primary_data)
    lonlat_column = lonlat_list
    df_way.insert(0, "lonlat", lonlat_column)

    way_linestring = map(lambda lonlats: LineString(lonlats), lonlat_list)
    length_column = (
        gpd.GeoSeries(way_linestring)
        .set_crs("EPSG:4326")
        .to_crs("EPSG:3857")
        .length
    )

    df_way.insert(0, "Length", length_column)

def convert_ways(df_way, primary_data):
    lonlat_list = lonlat_lookup(df_way, primary_data)



# Combine convert_pd_to_gdf_nodes and convert_pd_to_gdf_lines int0
# convert_pd_to_gdf
def convert_pd_to_gdf_nodes(df):

    # TODO: Ensure Type = node
    gdf = gpd.GeoDataFrame(
        df, geometry=[Point(x, y) for x, y in df.lonlat], crs="EPSG:4326"
    )
    gdf.drop(columns=["lonlat"], inplace=True)
    return gdf

def convert_pd_to_gdf_lines(df_way):
    # TODO: Ensure Type = way
    gdf = gpd.GeoDataFrame(
        df_way, geometry=[LineString(x) for x in df_way.lonlat], crs="EPSG:4326"
    )
    gdf.drop(columns=["lonlat"], inplace=True)
    return gdf

def get_region_slug(region_list):
    if len(region_list) == 1:
        region_slug = str(region_list[0].short)
    else:
        # TODO: Implement filenamer for multiple regions
        raise NotImplementedError

    return region_slug



def output_creation(df_feature, region_list, feature_name, data_dir, out_format):
    """
    Save Dataframe to disk
    Currently supports 
        CSV: Comma Separated Values
        GeoJSON: Seperate files for Nodes and Ways

    Args:
        df_feature:
    """

    region_slug = get_region_slug(region_list) # country code e.g. BJ
    out_dir = os.path.join(data_dir, "out")  # Output file directory
    out_slug = os.path.join(out_dir, f"{region_slug}_{feature_name}")
    

    if not os.path.exists(out_dir):
        os.makedirs(
            out_dir, exist_ok=True
        )  # create raw directory

    df_feature.reset_index(drop=True, inplace=True)

    # Generate Files
    if df_feature.empty:
        logger.warning(f"All feature data frame empty for {feature_name}")
        return None

    if "csv" in out_format:
        logger.debug("Writing CSV file")
        df_feature.to_csv(out_slug + '.csv')

    if "geojson" in out_format:
        logger.debug("Writing GeoJSON file")
        # if primary_feature_element[primary_name][feature_name] == "way":
        #     gdf_feature = convert_pd_to_gdf_lines(df_feature)
        # else:
        #     gdf_feature = convert_pd_to_gdf_nodes(df_feature)

        # try:
        #     gdf_feature.drop(columns=["refs"], inplace=True)
        # except:
        #     pass

        gdf_feature.to_file(out_slug + '.geojson', driver='GeoJSON')




__author__ = "PyPSA meets Earth"
__copyright__ = "Copyright 2022, The PyPSA meets Earth Initiative"
__license__ = "MIT"

"""
This module contains the config for osm features.

"""

# The primary_feature_element represents the final representation of the feature
# For node features: ways are converted to nodes
# For way features: only ways are used

# TODO: Add 'keep' to keep the original feature representation
#%%
# import json
from pprint import pprint
import pandas as pd
# Node (Towers)
# Way (Lines and Cables)
# Area (Substations and Generators)

#%%
# TODO: Add taginfo.json to package data
# TODO: Can be used to get the primary keys and values
# TODO: Implement view features simialr to gfk_data
# taginfo_url='https://raw.githubusercontent.com/openstreetmap/id-tagging-schema/main/dist/taginfo.min.json'
# taginfo = '/home/matin/earth-osm/taginfo.min.json'
# with open(taginfo, encoding='utf8') as f:
#     d = json.load(f)
# tag_dict = d['tags']
# pprint(tag_dict)
# tag_df = pd.DataFrame(tag_dict)

# #%%
# power_df = tag_df.loc[tag_df['key'] == 'power']

# #%%
# power_df.to_dict('records')

# #%%
# by_parent = tag_df.groupby("key")[["value"]].count()

# #%%
# tag_df.drop(['object_types', 'icon_url'], axis=1).value_counts()
# by_parent
# parent_dict = by_parent.groups


#%%
# primary_feature_element = {
#     "power": {
#         "substation": "node", #(node, area)
#         "generator": "node", # (node, way, area)
#         "line": "way", #(way)
#         "tower": "node", #(node)
#         "cable": "way", #(way)
#     }
#     # add more primary features here
# }


primary_feature_element = {
    "power": {
        "substation": ("node", "area"), 
        "generator": ("node", "way", "area"),
        "line": ("way"),
        "tower": ("node"),
        "cable": ("way"),
        }
    # add more primary features here
}


# ===============================
# OSM FEATURE COLUMNS
# ===============================
# These configurations are used to specify which OSM tags are kept as columns in DataFrame.
# Follows the OSM Wiki: https://wiki.openstreetmap.org/wiki/Power


# "Length" is added for way features
# "Area" is added for node features

# ========================
# BASIC INFO TAGS
# ========================
# A list of tags that are relevant for most OSM keys

columns_basic = [
    "id",
    "lonlat",
    "tags.power",
    "Type",
    "Country",
    #"refs"
]

# ========================
# SUBSTATION TAGS
# ========================

# Default tags to keep as columns with substation
# Based on: https://wiki.openstreetmap.org/wiki/Key:substation
columns_substation = [
    "Area",
    "tags.substation",
    "tags.voltage",
    # Other tags which are not kept by default
    # =====================================
    # "TODO:ADD Tags not kept here",
]

# ========================
# GENERATOR TAGS
# ========================

# Default tags to keep as columns with generator
# Based on: https://wiki.openstreetmap.org/wiki/Key:generator

columns_generator = [
    "Area",
    "tags.name",
    "tags.generator:type",
    "tags.generator:method",
    "tags.generator:source",
    "tags.generator:output:electricity",
    # Other tags which are not kept by default
    # =====================================
    # "TODO:ADD Tags not kept here",
]

# ========================
# LINE TAGS
# ========================

# Default tags to keep as columns with line
# Based on: https://wiki.openstreetmap.org/wiki/Key:line

columns_line = [
    "Length",
    "tags.cables",
    "tags.voltage",
    "tags.circuits",
    "tags.frequency",
    # Other tags which are not kept by default
    # =====================================
    # "TODO:ADD Tags not kept here",
]

# ========================
# CABLE TAGS
# ========================

# Default tags to keep as columns with substation
# Based on: https://wiki.openstreetmap.org/wiki/Key:cable

columns_cable = [
    "Length",
    "tags.cables",
    "tags.voltage",
    "tags.circuits",
    "tags.frequency",
    "tags.location",
    # Other tags which are not kept by default
    # =====================================
    # "TODO:ADD Tags not kept here",
]

# ========================
# TOWER TAGS
# ========================

# Default tags to keep as columns with tower
# Based on: https://wiki.openstreetmap.org/wiki/Key:tower

columns_tower = [
    "Area",
    "tags.tower",
    "tags.material",
    "tags.structure",
    "tags.operator",
    "tags.line_attachment",
    "tags.line_management",
    "tags.ref",
    "tags.height",
    # Other tags which are not kept by default
    # =====================================
    # "TODO:ADD Tags not kept here",
]

## FINAL DICTIONARY
feature_columns = {
    "substation": columns_basic + columns_substation,
    "generator": columns_basic + columns_generator,
    "line": columns_basic + columns_line,
    "cable": columns_basic + columns_cable,
    "tower": columns_basic + columns_tower,
}



"""Python setup.py for earth_osm package"""
import io
import os
from setuptools import find_packages, setup


def read(*paths, **kwargs):
    """Read the contents of a text file safely.
    >>> read("earth_osm", "VERSION")
    '0.1.0'
    >>> read("README.md")
    ...
    """
    content = ""
    with io.open(
        os.path.join(os.path.dirname(__file__), *paths),
        encoding=kwargs.get("encoding", "utf8"),
    ) as open_file:
        content = open_file.read().strip()
    return content


setup(
    name="earth_osm",
    version=read("earth_osm", "VERSION"),
    description="Python tool to extract large-amounts of OpenStreetMap data",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    author="pypsa-meets-earth",
    url="https://github.com/pypsa-meets-earth/earth-osm/",
    packages=find_packages(exclude=["docs", "tests"]),
    include_package_data=True,
    python_requires=">=3.6",
    entry_points={
        "console_scripts": ["earth_osm = earth_osm.__main__:main"]
    },
    install_requires=[
        "geopandas",
        "pandas",
        "tqdm",
        "requests",
        "protobuf>=4.21.1",
    ],
    extras_require={"test": [
        "pytest",
        "coverage",
        "flake8",
        "black",
        "isort",
        "pytest-cov",
        "codecov",
        "mypy>=0.9",
        "gitchangelog",
        "mkdocs",
        ],
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
    ],
)


__author__ = "PyPSA meets Earth"
__copyright__ = "Copyright 2022, The PyPSA meets Earth Initiative"
__license__ = "MIT"

"""Extract OSM from PBF Files

This module extracts OSM data from PBF files.

Modified from esy-osmfilter/pre_filter.py

"""

import itertools
import logging
import multiprocessing as mp

from earth_osm.osmpbf import Node, Relation, Way, osmformat_pb2
from earth_osm.osmpbf.file import iter_blocks, iter_primitive_block, read_blob

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def primary_entry_filter(entry, pre_filter):
    filtermap = pre_filter[type(entry)]  # {"power": ['line', 'tower]}
    for primary_name in filtermap.keys():  # power
        if (primary_name in entry.tags.keys()):  # tags a dict with keys (...'power')
            # for filtermap.get(key) is ['line', 'tower']
            #  is the value for power in tags
            # feature_name == True gets all things tagged with power
            return any(
                feature_name == True
                or feature_name in entry.tags.get(primary_name)
                for feature_name in filtermap.get(primary_name)
            )


def id_filter(entry, idset):
    return entry.id in idset


def way_filter(entry, way_relation_members):
    return type(entry) is Way and entry.id in way_relation_members


def filter_file_block(filename, ofs, header, filter_func, args, kwargs):
    with open(filename, "rb") as file:
        entries = osmformat_pb2.PrimitiveBlock()
        entries.ParseFromString(read_blob(file, ofs, header))
        return [
            entry
            for entry in iter_primitive_block(entries)
            if filter_func(entry, *args, **kwargs)
        ]


def pool_file_query(filename, pool):
    """
    returns query function that accepts a filter function and returns a list of filtered entries
    """
    with open(filename, "rb") as file:
        blocks = [(file.name, ofs, header) for ofs, header in iter_blocks(file)]

    def query_func(filter_func, *args, **kwargs):
        entry_lists = pool.starmap(
            filter_file_block,
            [block + (filter_func, args, kwargs) for block in blocks],
        )
        
        return itertools.chain(*(entries for entries in entry_lists))

    return query_func


def filter_pbf(filename, pre_filter, multiprocess=True):
    """
    Parallized pre-Filtering of OSM file by a pre_filter

    Args:
        filename:   PBF file
        pre_filter: dict of dicts in the following format: {
                Node: {primary_name: feature_list}, 
                Way: {primary_name: feature_list},
                Relation: {primary_name: feature_list}}

    Returns:
        targetname: JSON-file
    """

    with mp.Pool(processes=1 if not multiprocess else mp.cpu_count() - 1 or 1) as pool:
        file_query = pool_file_query(filename, pool)    
        primary_entries = list(file_query(primary_entry_filter, pre_filter)) #list of named  tuples eg. Node(id,tags, lonlat)
        
        node_relation_members = set()
        way_relation_members = set()
        way_refs = set()
        relation_way_node_members = set()
        
        for entry in primary_entries:
    
            if type(entry) is Relation:
                for id, typename, role in entry.members:
                    if typename == "NODE":
                        node_relation_members.add(id)
                    elif typename == "WAY":
                        way_relation_members.add(id)
            
            if type(entry) is Way:
                way_refs.update(entry.refs)

        way_entries = list(file_query(way_filter, way_relation_members))

        for entry in way_entries:
            relation_way_node_members.update(entry.refs)

        primary_entries.extend(
            file_query(
                id_filter,
                set.union(
                    node_relation_members,
                    way_refs,
                    way_relation_members,
                    relation_way_node_members,
                ),
            )
        )

        primary_entries.sort(key=lambda entry: entry.id)
        
        primary_data = {"Node": {}, "Way": {}, "Relation": {}}
        for entry in primary_entries:
            primary_data[type(entry).__name__][str(entry.id)] = dict(entry._asdict())
    
    return primary_data



from earth_osm.gfk_data import (
    get_root_list,
    get_all_regions_dict,
    view_regions,
    get_id_by_code,
    get_code_by_id,
    get_region_dict,
    get_id_by_str,
    get_region_tuple,
    get_all_valid_list
)


def test_view():
    from pprint import pprint
    print(get_root_list())
    pprint(get_all_regions_dict(level=2))
    view_regions()  # TODO: replace Nan with blank

def test_regions():
    # pprint(get_region_dict('germany')) #Raises KeyError
    assert set(get_region_dict("germany").keys()) == set(
        ["id", "name", "parent", "short_code", "urls"]
    )

    # print(get_id_by_code('DE')) #Supresses KeyError
    assert get_id_by_code("DE") == "germany"

    # print(get_code_by_id('germany')) #Supresses KeyError
    assert get_code_by_id("germany") == "DE"

    # print(get_id_by_str('germany')) #Raises KeyError
    assert get_id_by_str("germany") == "germany"
    assert get_id_by_str("DE") == "germany"
    assert get_id_by_str("DE") == "germany"
    try:
        get_id_by_str("unknown country")
    except KeyError:
        assert True

def test_region_tuple():
    # print(get_region_tuple('germany'))
    print(get_region_tuple("germany"))
    print(get_region_tuple("germany").short)
    print(get_all_valid_list())



__author__ = "PyPSA meets Earth"
__copyright__ = "Copyright 2022, The PyPSA meets Earth Initiative"
__license__ = "MIT"

"""
This is the principal module of the earth_osm project.
"""
#%%
import logging
import os

import pandas as pd

from earth_osm.config import primary_feature_element, feature_columns
from earth_osm.filter import get_filtered_data
from earth_osm.gfk_data import get_region_tuple
from earth_osm.utils import convert_ways_lines, convert_ways_points, output_creation


logger = logging.getLogger("osm_data_extractor")
logger.setLevel(logging.INFO)

#%%
# TODO: Rename to process_region
def process_country(region, primary_name, feature_name, mp, update, data_dir):
    """
    Process Country

    Args:
        region: Region object
        primary_name: Primary Feature Name
        feature_name: Feature Name
        mp: Multiprocessing object
        update: Update flag

    Returns:
        None
    """
    primary_dict, feature_dict = get_filtered_data(region, primary_name, feature_name, mp, update, data_dir)

    primary_data = primary_dict['Data']
    feature_data = feature_dict['Data']

    df_node = pd.json_normalize(feature_data["Node"].values())
    df_way = pd.json_normalize(feature_data["Way"].values())

    element_set = set(primary_feature_element[primary_name][feature_name])
    assert element_set <= set(['node', 'way', 'area'])

    if "node" in element_set:
        df_node["Type"] = "Node"
    
    if set(['way', 'area']) <= element_set:
        
    
    # element_type = primary_feature_element[primary_name][feature_name]

    # if element_type == "way":
    #     convert_ways_lines(
    #         df_way, primary_data
    #     ) if not df_way.empty else logger.warning(
    #         f"Empty Way Dataframe for {feature_name} in {region.short}"
    #     )
    #     if not df_node.empty:
    #         logger.warning(
    #             f"Node dataframe not empty for {feature_name} in {region.short}"
    #         )

    # if element_type == "node":
    #     convert_ways_points(df_way, primary_data) if not df_way.empty else None

    # Add Original Type Column
    
    df_way["Type"] = "Way"


    # Concatinate Nodes and Ways
    df_feature = pd.concat([df_node, df_way], axis=0)

    # Add Country Column
    # TODO: rename Country to Region
    # df_feature["Country"] = region.short

    return df_feature


def get_osm_data(
    region_list=['germany'],
    primary_name='power',
    feature_list=['tower'],
    update=False,
    mp=True,
    data_dir=os.path.join(os.getcwd(), 'earth_data'),
    out_format="csv",
    out_aggregate=False,
):
    """
    Get OSM Data for a list of regions and features
    args:
        region_list: list of regions to get data for
        primary_name: primary feature to get data for
        feature_list: list of features to get data for
        update: update data
        mp: use multiprocessing
    returns:
        dict of dataframes
    """
    region_tuple_list = [get_region_tuple(rs) for rs in region_list]

    for region in region_tuple_list:
        for feature_name in feature_list:
            df_feature = process_country(region, primary_name, feature_name, mp, update, data_dir)
            # df_feature = df_feature.reindex(columns=feature_columns[feature_name])
            
            output_creation(df_feature, primary_name, feature_name, [region], data_dir, out_format, out_aggregate)
